var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "@babel/types"], factory);
    }
})(function (require, exports) {
    "use strict";
    const t = __importStar(require("@babel/types"));
    const FUNCNAME = "__VueFC__";
    /**
     * Build object literal represents functional component.
     *
     * Input: `
     *   ctx => (<div>{ctx.props.text}</div>)
     * `
     * Output: `
     *   {
     *     functional: true,
     *     render(h, ctx) {
     *       return <div>{ctx.props.text}</div>;
     *     }
     *   }
     * `
     */
    function buildFunctionalComponentObject(renderFn, componentName) {
        let body;
        if (t.isBlockStatement(renderFn.body)) {
            body = renderFn.body;
        }
        else {
            body = t.blockStatement([t.returnStatement(renderFn.body)]);
        }
        const ret = t.objectExpression([
            t.objectProperty(t.identifier("functional"), t.booleanLiteral(true)),
            t.objectMethod("method", t.identifier("render"), [t.identifier("h"), ...renderFn.params], body)
        ]);
        if (componentName) {
            ret.properties.push(t.objectProperty(t.identifier("name"), t.stringLiteral(componentName)));
        }
        return ret;
    }
    /**
     * Get function name or method name from CallExpression.
     */
    function getCalleeName(node) {
        const callee = node.callee;
        if (t.isIdentifier(callee)) {
            // xxx(...)
            // ^^^
            return callee.name;
        }
        if (t.isMemberExpression(callee)) {
            const property = callee.property;
            if (t.isIdentifier(property)) {
                // xxx.yyy(...)
                //     ^^^
                return property.name;
            }
            if (t.isStringLiteral(property)) {
                // xxx["yyy"](...)
                //      ^^^
                return property.value;
            }
        }
        return undefined;
    }
    /**
     * Guess component name from __VueFC__ call expression.
     *
     * expected patterns:
     *  1) const ComponentName = xxx.__VueFC__(...);
     *  2) obj.ComponentName = xxx.__VueFC__(...);
     *  3) ComponentName = xxx.__VueFC__(...);
     *  4) export const ComponentName = xxx.__VueFC__(...);
     */
    function guessComponentName(path) {
        const parent = path.parentPath.node;
        let componentName;
        if (t.isAssignmentExpression(parent) && parent.right === path.node) {
            const left = parent.left;
            if (t.isIdentifier(left)) {
                componentName = left.name;
            }
            else if (t.isMemberExpression(left)) {
                componentName = t.isIdentifier(left.property)
                    ? left.property.name
                    : undefined;
            }
        }
        else if (t.isVariableDeclarator(parent)) {
            componentName = t.isIdentifier(parent.id) ? parent.id.name : undefined;
        }
        else if (t.isExportDefaultDeclaration(parent)) {
            // no name
        }
        else {
            throw path.buildCodeFrameError(`result of ${FUNCNAME} must be assigned or exported`);
        }
        if (componentName && /^[A-Z]/.test(componentName)) {
            return componentName;
        }
        else {
            return undefined;
        }
    }
    /**
     * Replace __VueFC__ call expression by object literal which represents
     * functional component.
     */
    function processCallExpression(path) {
        const { node } = path;
        const funcname = getCalleeName(node);
        if (funcname !== FUNCNAME) {
            return;
        }
        if (node.arguments.length !== 1) {
            throw path.buildCodeFrameError(`${FUNCNAME} must have exactly 1 argument`);
        }
        const arg = node.arguments[0];
        if (!t.isArrowFunctionExpression(arg)) {
            throw path.buildCodeFrameError(`argument of ${FUNCNAME} must be arrow function`);
        }
        const componentName = guessComponentName(path);
        const functionalComponent = buildFunctionalComponentObject(arg, componentName);
        path.replaceWith(functionalComponent);
    }
    return function () {
        const plugin = {
            name: "vue-tsx-functional",
            visitor: {
                Program(path) {
                    path.traverse({
                        CallExpression(path) {
                            processCallExpression(path);
                        }
                    });
                }
            }
        };
        return plugin;
    };
});
